from random import random, shuffle
from matplotlib import pyplot as plt
from time import time

def distance(x, y):
    """
    compute the Euclidean distance between two 2d points x and y

    Arguments:
        x (Tuple[float, float]): the first point
        y (Tuple[float, float]): the second point

    Returns:
        float: the euclidean distance between x and y
    """
    return ((x[0] - y[0]) ** 2 + (x[1] - y[1]) ** 2) ** 0.5

def random_locations(n):
    """
    generate n random 2d points

    Arguments:
        n (int): the number of points to generate

    Returns:
        List[Tuple[float, float]]: a list of x, y coordinates represented as tuples
    """
    points = []
    for i in range(n):
        points.append((random(), random()))
    return points

def distance_matrix(points):
    """
    compute the matrix of euclidean distances between all pairs of points passed in a list. 
    The matrix is represented as a list of lists, so the distance between points i and j 
    is dists[i][j]

    Arguments:
        points (List[Tuple[float, float]]): a list of points 
            ((x, y) coordinates represented as tuples)

    Returns:
        List[List[float]]: a matrix representing the distances between pairs of points
    """
    n = len(points)
    distances = []
    for i in range(n):
        distances.append([])
        for j in range(n):
            # add the distance to the end of the last list
            distances[-1].append(distance(points[i], points[j]))

    return distances


def plot_tour(seq, points, title):
    """
    Visualize a path through a sequence of points and save it to the img directory. 

    Arguments:
        seq (List[int]): the order that the points are visited in
        points (List[Tuple[float, float]]): a list of points 
            ((x, y) coordinates represented as tuples)
        title (str): the intended title of the plot

    Returns: None
    """
    n = len(points)

    # create separate x and y lists for scatter plot
    x = []
    for i in range(n):
        x.append(points[seq[i]][0])
    x.append(points[seq[0]][0])
    
    y = []
    for i in range(n):
        y.append(points[seq[i]][1])
    y.append(points[seq[0]][1])

    # create plot
    plt.plot(x, y)
    plt.scatter(x, y, color="black", zorder=100)
    plt.scatter(x[0], y[0], color="orange", zorder=100)
    
    # compute distance for title
    dists = distance_matrix(points)
    n = len(seq)
    dist = 0
    for i in range(n - 1):
        dist += dists[seq[i]][seq[i + 1]]
    dist += dists[seq[n - 1]][seq[0]]

    # set title and axis bounds
    plt.title(f"{title}\nDistance traveled: {dist:.2f}km")
    plt.gca().set_xlim(0, 1)
    plt.gca().set_ylim(0, 1)
    
    # save and close figure
    plt.savefig(f"img/{title}.png")
    plt.close()


def plot_times(times, title):
    """
    Visualize a series of timing results and save them to the img directory. 

    Arguments:
        times (List[float]): a list of runtimes in seconds
        title (str): the title of the plot

    Returns: None
    """
    # create plot
    plt.scatter(list(range(2, len(times) + 2)), times)
    plt.title(f"{title}")
    plt.gca().set_xlabel("Number of locations to visit")
    plt.gca().set_ylabel("Time elapsed (seconds)")

    # save to png file
    plt.savefig(f"img/{title}.png")
    plt.close()
    

def random_tour(points):
    """
    create a random sequence of indices for points.

    Arguments:
        points (List[Tuple[float, float]]): a list of points 
            ((x, y) coordinates represented as tuples)
        
    Returns:
        List[int]: a list of indices representing a random sequence of points
    """
    s = list(range(len(points)))
    shuffle(s)
    return s    

def time_search(search_fun, n):
    """
    time the result of running a search function on a problem of size n. 

    Arguments:
        search_fun (Callable): the function to call
            NOTE for 145 students: you don't need to know how to use a function as an argument.
        n (int): the number of points in the tsp

    Returns:
        float: the number of seconds taken
    """
    points = random_locations(n)
    dists = distance_matrix(points)
    
    start = time()
    # these aren't actually used, we just need to get the time
    seq, dist = search_fun(dists)
    end = time()
    
    return end - start
